@@ -1,9 +1,7 @@
 /**
  * PyAudio: Python Bindings for PortAudio.
  *
- * May 2006: Supports Non-Blocking mode only
- *
- * Copyright (c) 2006-2008 Hubert Pham
+ * Copyright (c) 2006-2012 Hubert Pham
  *
  * Permission is hereby granted, free of charge, to any person
  * obtaining a copy of this software and associated documentation
@@ -38,6 +36,10 @@
 #define DEFAULT_FRAMES_PER_BUFFER 1024
 /* #define VERBOSE */
 
+#define min(a,b) \
+   ({ __typeof__ (a) _a = (a); \
+       __typeof__ (b) _b = (b); \
+     _a < _b ? _a : _b; })
 
 /************************************************************
  *
@@ -49,7 +51,6 @@
  *     - PaHostInfo
  *     - PaStream
  * III. PortAudio Method Implementations
- *      (BLOCKING MODE ONLY!)
  *     - Initialization/Termination
  *     - HostAPI
  *     - DeviceAPI
@@ -180,7 +181,7 @@ _pyAudio_paDeviceInfo_get_structVersion(_pyAudio_paDeviceInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->devInfo->structVersion);
+  return PyLong_FromLong(self->devInfo->structVersion);
 }
 
 static PyObject *
@@ -194,7 +195,7 @@ _pyAudio_paDeviceInfo_get_name(_pyAudio_paDeviceInfo *self,
     return NULL;
   }
 
-  return PyString_FromString(self->devInfo->name);
+  return PyUnicode_FromString(self->devInfo->name);
 }
 
 static PyObject *
@@ -208,7 +209,7 @@ _pyAudio_paDeviceInfo_get_hostApi(_pyAudio_paDeviceInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->devInfo->hostApi);
+  return PyLong_FromLong(self->devInfo->hostApi);
 }
 
 static PyObject *
@@ -222,7 +223,7 @@ _pyAudio_paDeviceInfo_get_maxInputChannels(_pyAudio_paDeviceInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->devInfo->maxInputChannels);
+  return PyLong_FromLong(self->devInfo->maxInputChannels);
 }
 
 static PyObject *
@@ -236,7 +237,7 @@ _pyAudio_paDeviceInfo_get_maxOutputChannels(_pyAudio_paDeviceInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->devInfo->maxOutputChannels);
+  return PyLong_FromLong(self->devInfo->maxOutputChannels);
 }
 
 static PyObject *
@@ -394,12 +395,11 @@ _pyAudio_paDeviceInfo_dealloc(_pyAudio_paDeviceInfo* self)
   self->devInfo = NULL;
 
   /* free the object */
-  self->ob_type->tp_free((PyObject*) self);
+  Py_TYPE(self)->tp_free((PyObject*) self);
 }
 
 static PyTypeObject _pyAudio_paDeviceInfoType = {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "_portaudio.paDeviceInfo", /*tp_name*/
     sizeof(_pyAudio_paDeviceInfo),   /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -478,7 +478,7 @@ _pyAudio_paHostApiInfo_get_structVersion(_pyAudio_paHostApiInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->apiInfo->structVersion);
+  return PyLong_FromLong(self->apiInfo->structVersion);
 }
 
 static PyObject *
@@ -492,7 +492,7 @@ _pyAudio_paHostApiInfo_get_type(_pyAudio_paHostApiInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong((long) self->apiInfo->type);
+  return PyLong_FromLong((long) self->apiInfo->type);
 }
 
 static PyObject *
@@ -506,7 +506,7 @@ _pyAudio_paHostApiInfo_get_name(_pyAudio_paHostApiInfo *self,
     return NULL;
   }
 
-  return PyString_FromString(self->apiInfo->name);
+  return PyUnicode_FromString(self->apiInfo->name);
 }
 
 static PyObject *
@@ -520,7 +520,7 @@ _pyAudio_paHostApiInfo_get_deviceCount(_pyAudio_paHostApiInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->apiInfo->deviceCount);
+  return PyLong_FromLong(self->apiInfo->deviceCount);
 }
 
 static PyObject *
@@ -534,7 +534,7 @@ _pyAudio_paHostApiInfo_get_defaultInputDevice(_pyAudio_paHostApiInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->apiInfo->defaultInputDevice);
+  return PyLong_FromLong(self->apiInfo->defaultInputDevice);
 }
 
 static PyObject *
@@ -548,7 +548,7 @@ _pyAudio_paHostApiInfo_get_defaultOutputDevice(_pyAudio_paHostApiInfo *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->apiInfo->defaultOutputDevice);
+  return PyLong_FromLong(self->apiInfo->defaultOutputDevice);
 }
 
 static int
@@ -569,7 +569,7 @@ _pyAudio_paHostApiInfo_dealloc(_pyAudio_paHostApiInfo* self)
   self->apiInfo = NULL;
 
   /* free the object */
-  self->ob_type->tp_free((PyObject*) self);
+  Py_TYPE(self)->tp_free((PyObject*) self);
 }
 
 static PyGetSetDef _pyAudio_paHostApiInfo_getseters[] = {
@@ -613,8 +613,7 @@ static PyGetSetDef _pyAudio_paHostApiInfo_getseters[] = {
 };
 
 static PyTypeObject _pyAudio_paHostApiInfoType = {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "_portaudio.paHostApiInfo", /*tp_name*/
     sizeof(_pyAudio_paHostApiInfo),   /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -694,6 +693,7 @@ _pyAudio_MacOSX_hostApiSpecificStreamInfo_cleanup(_pyAudio_Mac_HASSI *self)
 
   if (self->channelMap != NULL) {
     free(self->channelMap);
+    self->channelMap = NULL;
   }
 
   self->flags = paMacCorePlayNice;
@@ -704,14 +704,15 @@ static void
 _pyAudio_MacOSX_hostApiSpecificStreamInfo_dealloc(_pyAudio_Mac_HASSI *self)
 {
   _pyAudio_MacOSX_hostApiSpecificStreamInfo_cleanup(self);
-  self->ob_type->tp_free((PyObject *) self);
+  Py_TYPE(self)->tp_free((PyObject *) self);
 }
 
 static int
-_pyAudio_MacOSX_hostApiSpecificStreamInfo_init(_pyAudio_Mac_HASSI *self,
+_pyAudio_MacOSX_hostApiSpecificStreamInfo_init(PyObject *_self,
 					       PyObject *args,
 					       PyObject *kwargs)
 {
+  _pyAudio_Mac_HASSI *self = (_pyAudio_Mac_HASSI *) _self;
   PyObject *channel_map = NULL;
   int flags = paMacCorePlayNice;
 
@@ -756,15 +757,18 @@ _pyAudio_MacOSX_hostApiSpecificStreamInfo_init(_pyAudio_Mac_HASSI *self,
       }
 
       // make sure element is an integer
-      if (!PyInt_Check(element)) {
+      if (!PyNumber_Check(element)) {
 	PyErr_SetString(PyExc_ValueError,
 			"Channel Map must consist of integer elements");
 	_pyAudio_MacOSX_hostApiSpecificStreamInfo_cleanup(self);
 	return -1;
       }
 
+      PyObject *long_element = PyNumber_Long(element);
+
       // OK, looks good
-      self->channelMap[i] = (SInt32) PyInt_AsLong(element);
+      self->channelMap[i] = (SInt32) PyLong_AsLong(long_element);
+      Py_DECREF(long_element);
     }
   }
 
@@ -795,7 +799,7 @@ static PyObject *
 _pyAudio_MacOSX_hostApiSpecificStreamInfo_get_flags(_pyAudio_Mac_HASSI *self,
 						    void *closure)
 {
-  return PyInt_FromLong(self->flags);
+  return PyLong_FromLong(self->flags);
 }
 
 static PyObject *
@@ -811,7 +815,7 @@ _pyAudio_MacOSX_hostApiSpecificStreamInfo_get_channel_map(
   int i;
   PyObject *channelMapTuple = PyTuple_New(self->channelMapSize);
   for (i = 0; i < self->channelMapSize; ++i) {
-    PyObject *element = PyInt_FromLong(self->channelMap[i]);
+    PyObject *element = PyLong_FromLong(self->channelMap[i]);
     if (!element) {
       PyErr_SetString(PyExc_SystemError, "Invalid channel map");
       return NULL;
@@ -819,7 +823,7 @@ _pyAudio_MacOSX_hostApiSpecificStreamInfo_get_channel_map(
 
     if (PyTuple_SetItem(channelMapTuple,
 			i,
-			PyInt_FromLong(self->channelMap[i]))) {
+			PyLong_FromLong(self->channelMap[i]))) {
       // non-zero on error
       PyErr_SetString(PyExc_SystemError, "Can't create channel map.");
       return NULL;
@@ -856,8 +860,7 @@ static PyGetSetDef _pyAudio_MacOSX_hostApiSpecificStreamInfo_getseters[] = {
 };
 
 static PyTypeObject _pyAudio_MacOSX_hostApiSpecificStreamInfoType = {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "_portaudio.PaMacCoreStreamInfo", /*tp_name*/
     sizeof(_pyAudio_MacOSX_hostApiSpecificStreamInfo),   /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -893,7 +896,7 @@ static PyTypeObject _pyAudio_MacOSX_hostApiSpecificStreamInfoType = {
     0,  /* tp_descr_get */
     0,  /* tp_descr_set */
     0,  /* tp_dictoffset */
-    _pyAudio_MacOSX_hostApiSpecificStreamInfo_init,  /* tp_init */
+    (int (*)(PyObject*, PyObject*, PyObject*))_pyAudio_MacOSX_hostApiSpecificStreamInfo_init,  /* tp_init */
     0,  /* tp_alloc */
     0,  /* tp_new */
 };
@@ -904,6 +907,12 @@ static PyTypeObject _pyAudio_MacOSX_hostApiSpecificStreamInfoType = {
  * Stream Wrapper Python Object
  *************************************************************/
 
+typedef struct {
+  PyObject *callback;
+  long main_thread_id;
+  unsigned int frame_size;
+} PyAudioCallbackContext;
+
 typedef struct {
   PyObject_HEAD
   PaStream *stream;
@@ -913,6 +922,9 @@ typedef struct {
   /* include PaStreamInfo too! */
   PaStreamInfo *streamInfo;
 
+  /* context for callback */
+  PyAudioCallbackContext *callbackContext;
+
   int is_open;
 } _pyAudio_Stream;
 
@@ -942,6 +954,12 @@ _cleanup_Stream_object(_pyAudio_Stream *streamObject)
     streamObject->outputParameters = NULL;
   }
 
+  if (streamObject->callbackContext != NULL) {
+    Py_XDECREF(streamObject->callbackContext->callback);
+    free(streamObject->callbackContext);
+    streamObject->callbackContext = NULL;
+  }
+
   /* designate the stream as closed */
   streamObject->is_open = 0;
 }
@@ -953,7 +971,7 @@ _pyAudio_Stream_dealloc(_pyAudio_Stream* self)
   _cleanup_Stream_object(self);
 
   /* free the object */
-  self->ob_type->tp_free((PyObject*) self);
+  Py_TYPE(self)->tp_free((PyObject*) self);
 }
 
 
@@ -978,7 +996,7 @@ _pyAudio_Stream_get_structVersion(_pyAudio_Stream *self,
     return NULL;
   }
 
-  return PyInt_FromLong(self->streamInfo->structVersion);
+  return PyLong_FromLong(self->streamInfo->structVersion);
 }
 
 static PyObject *
@@ -1096,8 +1114,7 @@ static PyGetSetDef _pyAudio_Stream_getseters[] = {
 };
 
 static PyTypeObject _pyAudio_StreamType = {
-    PyObject_HEAD_INIT(NULL)
-    0,                         /*ob_size*/
+    PyVarObject_HEAD_INIT(NULL, 0)
     "_portaudio.Stream",       /*tp_name*/
     sizeof(_pyAudio_Stream),         /*tp_basicsize*/
     0,                         /*tp_itemsize*/
@@ -1165,7 +1182,7 @@ pa_get_version(PyObject *self, PyObject *args)
   if (!PyArg_ParseTuple(args, ""))
     return NULL;
 
-  return PyInt_FromLong(Pa_GetVersion());
+  return PyLong_FromLong(Pa_GetVersion());
 }
 
 static PyObject *
@@ -1174,7 +1191,7 @@ pa_get_version_text(PyObject *self, PyObject *args)
   if (!PyArg_ParseTuple(args, ""))
     return NULL;
 
-  return PyString_FromString(Pa_GetVersionText());
+  return PyUnicode_FromString(Pa_GetVersionText());
 }
 
 /*************************************************************
@@ -1239,7 +1256,7 @@ pa_get_host_api_count(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(count);
+  return PyLong_FromLong(count);
 }
 
 static PyObject *
@@ -1266,7 +1283,7 @@ pa_get_default_host_api(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(index);
+  return PyLong_FromLong(index);
 }
 
 static PyObject *
@@ -1294,7 +1311,7 @@ pa_host_api_type_id_to_host_api_index(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(index);
+  return PyLong_FromLong(index);
 }
 
 static PyObject *
@@ -1323,7 +1340,7 @@ pa_host_api_device_index_to_device_index(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(devIndex);
+  return PyLong_FromLong(devIndex);
 }
 
 static PyObject *
@@ -1352,8 +1369,6 @@ pa_get_host_api_info(PyObject *self, PyObject *args)
   return (PyObject *) py_info;
 }
 
-
-
 /*************************************************************
  * Device API
  *************************************************************/
@@ -1381,7 +1396,7 @@ pa_get_device_count(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(count);
+  return PyLong_FromLong(count);
 }
 
 static PyObject *
@@ -1410,7 +1425,7 @@ pa_get_default_input_device(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(index);
+  return PyLong_FromLong(index);
 }
 
 static PyObject *
@@ -1439,7 +1454,7 @@ pa_get_default_output_device(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(index);
+  return PyLong_FromLong(index);
 }
 
 static PyObject *
@@ -1470,16 +1485,173 @@ pa_get_device_info(PyObject *self, PyObject *args)
  * Stream Open / Close / Supported
  *************************************************************/
 
+int
+_stream_callback_cfunction(const void *input,
+                           void *output,
+                           unsigned long frameCount,
+                           const PaStreamCallbackTimeInfo *timeInfo,
+                           PaStreamCallbackFlags statusFlags,
+                           void *userData)
+{
+  int return_val = paAbort;
+  PyGILState_STATE _state = PyGILState_Ensure();
+
+#ifdef VERBOSE
+  if (statusFlags != 0) {
+    printf("Status flag set: ");
+    if (statusFlags & paInputUnderflow) {
+      printf("input underflow!\n");
+    }
+    if (statusFlags & paInputOverflow) {
+      printf("input overflow!\n");
+    }
+    if (statusFlags & paOutputUnderflow) {
+      printf("output underflow!\n");
+    }
+    if (statusFlags & paOutputUnderflow) {
+      printf("output overflow!\n");
+    }
+    if (statusFlags & paPrimingOutput) {
+      printf("priming output!\n");
+    }
+  }
+#endif
+
+  PyAudioCallbackContext *context = (PyAudioCallbackContext *)userData;
+  PyObject *py_callback = context->callback;
+  unsigned int bytes_per_frame = context->frame_size;
+  long main_thread_id = context->main_thread_id;
+
+  PyObject *py_frame_count = PyLong_FromUnsignedLong(frameCount);
+  PyObject *py_time_info = Py_BuildValue("{s:d,s:d,s:d}",
+                                         "input_buffer_adc_time",
+                                         timeInfo->inputBufferAdcTime,
+                                         "current_time",
+                                         timeInfo->currentTime,
+                                         "output_buffer_dac_time",
+                                         timeInfo->outputBufferDacTime);
+  PyObject *py_status_flags = PyLong_FromUnsignedLong(statusFlags);
+  PyObject *py_input_data = Py_None;
+
+  if (input) {
+    py_input_data = PyBytes_FromStringAndSize(input,
+                                              bytes_per_frame * frameCount);
+  }
+
+  PyObject *py_result;
+  py_result = PyObject_CallFunctionObjArgs(py_callback,
+                                           py_input_data,
+                                           py_frame_count,
+                                           py_time_info,
+                                           py_status_flags,
+                                           NULL);
+
+  if (py_result == NULL) {
+#ifdef VERBOSE
+    fprintf(stderr, "An error occured while using the portaudio stream\n");
+    fprintf(stderr, "Error message: Could not call callback function\n");
+#endif
+    PyObject *err = PyErr_Occurred();
+
+    if (err) {
+        PyThreadState_SetAsyncExc(main_thread_id, err);
+
+        // Print out a stack trace to help debugging.
+        // TODO: make VERBOSE a runtime flag so users can control
+        // the amount of logging.
+        PyErr_Print();
+    }
+
+    goto end;
+  }
+
+  const char *pData;
+  int output_len;
+
+  if (!PyArg_ParseTuple(py_result,
+                        "z#i",
+                        &pData,
+                        &output_len,
+                        &return_val)) {
+#ifdef VERBOSE
+    fprintf(stderr, "An error occured while using the portaudio stream\n");
+    fprintf(stderr, "Error message: Could not parse callback return value\n");
+#endif
+
+    PyObject *err = PyErr_Occurred();
+
+    if (err) {
+        PyThreadState_SetAsyncExc(main_thread_id, err);
+
+        // Print out a stack trace to help debugging.
+        // TODO: make VERBOSE a runtime flag so users can control
+        // the amount of logging.
+        PyErr_Print();
+    }
+
+    Py_XDECREF(py_result);
+    return_val = paAbort;
+    goto end;
+  }
+
+  Py_DECREF(py_result);
+
+  if ((return_val != paComplete) &&
+      (return_val != paAbort) &&
+      (return_val != paContinue)) {
+      PyErr_SetString(PyExc_ValueError,
+                      "Invalid PaStreamCallbackResult from callback");
+      PyThreadState_SetAsyncExc(main_thread_id, PyErr_Occurred());
+      PyErr_Print();
+
+      // Quit the callback loop
+      return_val = paAbort;
+
+      goto end;
+  }
+
+  // Copy bytes for playback only if this is an output stream:
+
+  if (output) {
+      char *output_data = (char*)output;
+      memcpy(output_data, pData, min(output_len, bytes_per_frame * frameCount));
+
+      // Pad out the rest of the buffer with 0s if callback returned
+      // too few frames (and assume paComplete).
+      if (output_len < (frameCount * bytes_per_frame)) {
+          memset(output_data + output_len,
+                 0,
+                 (frameCount * bytes_per_frame) - output_len);
+          return_val = paComplete;
+      }
+  }
+
+ end:
+
+  if (input) {
+    // Decrement this at the end, after memcpy, in case the user
+    // returns py_input_data back for playback.
+    Py_DECREF(py_input_data);
+  }
+
+  Py_XDECREF(py_frame_count);
+  Py_XDECREF(py_time_info);
+  Py_XDECREF(py_status_flags);
+
+  PyGILState_Release(_state);
+  return return_val;
+}
+
 static PyObject *
 pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
 {
-
   int rate, channels;
   int input, output, frames_per_buffer;
   int input_device_index = -1;
   int output_device_index = -1;
   PyObject *input_device_index_arg = NULL;
   PyObject *output_device_index_arg = NULL;
+  PyObject *stream_callback = NULL;
   PaSampleFormat format;
   PaError err;
 
@@ -1510,13 +1682,14 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
 			   "frames_per_buffer",
 			   "input_host_api_specific_stream_info",
 			   "output_host_api_specific_stream_info",
+                           "stream_callback",
 			   NULL};
 
   if (!PyArg_ParseTupleAndKeywords(args, kwargs,
 #ifdef MACOSX
-				   "iik|iiOOiO!O!",
+				   "iik|iiOOiO!O!O",
 #else
-				   "iik|iiOOiOO",
+				   "iik|iiOOiOOO",
 #endif
 				   kwlist,
 				   &rate, &channels, &format,
@@ -1531,10 +1704,15 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
 #ifdef MACOSX
 				   &_pyAudio_MacOSX_hostApiSpecificStreamInfoType,
 #endif
-				   &outputHostSpecificStreamInfo))
+				   &outputHostSpecificStreamInfo,
+                                   &stream_callback))
 
     return NULL;
 
+  if (stream_callback && (PyCallable_Check(stream_callback) == 0)) {
+      PyErr_SetString(PyExc_TypeError, "stream_callback must be callable");
+      return NULL;
+  }
 
   /* check to see if device indices were specified */
   if ((input_device_index_arg == NULL) ||
@@ -1547,19 +1725,22 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
     input_device_index = -1;
 
   } else {
-
-    if (!PyInt_Check(input_device_index_arg)) {
+    // Support both Python 2 and Python 3 by using PyNumber_Check
+    if (!PyNumber_Check(input_device_index_arg)) {
       PyErr_SetString(PyExc_ValueError,
 		      "input_device_index must be integer (or None)");
       return NULL;
     }
 
-    input_device_index = PyInt_AsLong(input_device_index_arg);
+    PyObject *input_device_index_long =
+      PyNumber_Long(input_device_index_arg);
+
+    input_device_index = (int) PyLong_AsLong(input_device_index_long);
+    Py_DECREF(input_device_index_long);
 
 #ifdef VERBOSE
     printf("Using input device index number: %d\n", input_device_index);
 #endif
-
   }
 
   if ((output_device_index_arg == NULL) ||
@@ -1572,19 +1753,21 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
     output_device_index = -1;
 
   } else {
-
-    if (!PyInt_Check(output_device_index_arg)) {
+    // Support both Python 2 and Python 3 by using PyNumber_Check
+    if (!PyNumber_Check(output_device_index_arg)) {
       PyErr_SetString(PyExc_ValueError,
 		      "output_device_index must be integer (or None)");
       return NULL;
     }
 
-    output_device_index = PyInt_AsLong(output_device_index_arg);
+    PyObject *output_device_index_long =
+      PyNumber_Long(output_device_index_arg);
+    output_device_index = (int) PyLong_AsLong(output_device_index_long);
+    Py_DECREF(output_device_index_long);
 
 #ifdef VERBOSE
     printf("Using output device index number: %d\n", output_device_index);
 #endif
-
   }
 
   /* sanity checks */
@@ -1613,7 +1796,8 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
       outputParameters->device = output_device_index;
 
     /* final check -- ensure that there is a default device */
-    if (outputParameters->device < 0) {
+    if (outputParameters->device < 0 ||
+        outputParameters->device >= Pa_GetDeviceCount()) {
       free(outputParameters);
       PyErr_SetObject(PyExc_IOError,
 		      Py_BuildValue("(s,i)",
@@ -1677,6 +1861,16 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
 
   PaStream *stream = NULL;
   PaStreamInfo *streamInfo = NULL;
+  PyAudioCallbackContext *context = NULL;
+
+  // Handle callback mode:
+  if (stream_callback) {
+    Py_INCREF(stream_callback);
+    context = (PyAudioCallbackContext *) malloc(sizeof(PyAudioCallbackContext));
+    context->callback = (PyObject *) stream_callback;
+    context->main_thread_id = PyThreadState_Get()->thread_id;
+    context->frame_size = Pa_GetSampleSize(format) * channels;
+  }
 
   err = Pa_OpenStream(&stream,
 		      /* input/output parameters */
@@ -1690,10 +1884,10 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
 		      /* we won't output out of range samples
 			 so don't bother clipping them */
 		      paClipOff,
-		      /* no callback, use blocking API */
-		      NULL,
-		      /* no callback, so no callback userData */
-		      NULL);
+		      /* callback, if specified */
+		      (stream_callback)?(_stream_callback_cfunction):(NULL),
+		      /* callback userData, if applicable */
+		      context);
 
   if (err != paNoError) {
 
@@ -1725,6 +1919,7 @@ pa_open(PyObject *self, PyObject *args, PyObject *kwargs)
   streamObject->outputParameters = outputParameters;
   streamObject->is_open = 1;
   streamObject->streamInfo = streamInfo;
+  streamObject->callbackContext = context;
 
   return (PyObject *) streamObject;
 }
@@ -1765,7 +1960,7 @@ pa_get_sample_size(PyObject *self, PyObject *args)
     return NULL;
   }
 
-  return PyInt_FromLong(size_in_bytes);
+  return PyLong_FromLong(size_in_bytes);
 }
 
 
@@ -2238,8 +2433,8 @@ pa_read_stream(PyObject *self, PyObject *args)
   fprintf(stderr, "Allocating %d bytes\n", num_bytes);
 #endif
 
-  rv = PyString_FromStringAndSize(NULL, num_bytes);
-  sampleBlock = (short *) PyString_AsString(rv);
+  rv = PyBytes_FromStringAndSize(NULL, num_bytes);
+  sampleBlock = (short *) PyBytes_AsString(rv);
 
   if (sampleBlock == NULL) {
     PyErr_SetObject(PyExc_IOError,
@@ -2307,7 +2502,7 @@ pa_get_stream_write_available(PyObject *self, PyObject *args)
 
   PaStream *stream = streamObject->stream;
   frames = Pa_GetStreamWriteAvailable(stream);
-  return PyInt_FromLong(frames);
+  return PyLong_FromLong(frames);
 }
 
 static PyObject *
@@ -2332,7 +2527,7 @@ pa_get_stream_read_available(PyObject *self, PyObject *args)
 
   PaStream *stream = streamObject->stream;
   frames = Pa_GetStreamReadAvailable(stream);
-  return PyInt_FromLong(frames);
+  return PyLong_FromLong(frames);
 }
 
 
@@ -2342,30 +2537,60 @@ pa_get_stream_read_available(PyObject *self, PyObject *args)
  *
  ************************************************************/
 
+#if PY_MAJOR_VERSION >= 3
+#define ERROR_INIT NULL
+#else
+#define ERROR_INIT /**/
+#endif
+
+#if PY_MAJOR_VERSION >= 3
+static struct PyModuleDef moduledef = {
+  PyModuleDef_HEAD_INIT,
+  "_portaudio",
+  NULL,
+  -1,
+  paMethods,
+  NULL,
+  NULL,
+  NULL,
+  NULL
+};
+#endif
+
 PyMODINIT_FUNC
+#if PY_MAJOR_VERSION >= 3
+PyInit__portaudio(void)
+#else
 init_portaudio(void)
+#endif
 {
   PyObject* m;
 
+  PyEval_InitThreads();
+
   _pyAudio_StreamType.tp_new = PyType_GenericNew;
   if (PyType_Ready(&_pyAudio_StreamType) < 0)
-    return;
+    return ERROR_INIT;
 
   _pyAudio_paDeviceInfoType.tp_new = PyType_GenericNew;
   if (PyType_Ready(&_pyAudio_paDeviceInfoType) < 0)
-    return;
+    return ERROR_INIT;
 
   _pyAudio_paHostApiInfoType.tp_new = PyType_GenericNew;
   if (PyType_Ready(&_pyAudio_paHostApiInfoType) < 0)
-    return;
+    return ERROR_INIT;
 
 #ifdef MACOSX
   _pyAudio_MacOSX_hostApiSpecificStreamInfoType.tp_new = PyType_GenericNew;
   if (PyType_Ready(&_pyAudio_MacOSX_hostApiSpecificStreamInfoType) < 0)
-    return;
+    return ERROR_INIT;
 #endif
 
+#if PY_MAJOR_VERSION >= 3
+  m = PyModule_Create(&moduledef);
+#else
   m = Py_InitModule("_portaudio", paMethods);
+#endif
 
   Py_INCREF(&_pyAudio_StreamType);
   Py_INCREF(&_pyAudio_paDeviceInfoType);
@@ -2448,6 +2673,18 @@ init_portaudio(void)
   PyModule_AddIntConstant(m, "paIncompatibleStreamHostApi",
 			  paIncompatibleStreamHostApi);
 
+  /* callback constants */
+  PyModule_AddIntConstant(m, "paContinue", paContinue);
+  PyModule_AddIntConstant(m, "paComplete", paComplete);
+  PyModule_AddIntConstant(m, "paAbort", paAbort);
+
+  /* callback status flags */
+  PyModule_AddIntConstant(m, "paInputUnderflow", paInputUnderflow);
+  PyModule_AddIntConstant(m, "paInputOverflow", paInputOverflow);
+  PyModule_AddIntConstant(m, "paOutputUnderflow", paOutputUnderflow);
+  PyModule_AddIntConstant(m, "paOutputOverflow", paOutputOverflow);
+  PyModule_AddIntConstant(m, "paPrimingOutput", paPrimingOutput);
+
 #ifdef MACOSX
   PyModule_AddIntConstant(m, "paMacCoreChangeDeviceParameters",
 			  paMacCoreChangeDeviceParameters);
@@ -2472,4 +2709,8 @@ init_portaudio(void)
   PyModule_AddIntConstant(m, "paMacCoreMinimizeCPU",
 			  paMacCoreMinimizeCPU);
 #endif
+
+#if PY_MAJOR_VERSION >= 3
+  return m;
+#endif
 }
